@@ -2,6 +2,7 @@ import numpy as np
 import pandas as pd
 import warnings
 import os
+import shutil
 import tempfile
 import itertools
 import pathos
@@ -11,16 +12,30 @@ from inferelator_prior.processor.bedtools import (extract_bed_sequence, intersec
                                                   BED_CHROMOSOME)
 from inferelator_prior.processor.gtf import get_fasta_lengths, check_chromosomes_match
 
-INFO_COL = "Information Content"
-ENTROPY_COL = "Shannon Entropy"
+INFO_COL = "Information_Content"
+ENTROPY_COL = "Shannon_Entropy"
 OCC_COL = "Occurrence"
 LEN_COL = "Length"
 MOTIF_COL = "Motif_ID"
 MOTIF_NAME_COL = "Motif_Name"
+MOTIF_OBJ_COL = "Motif_Object"
+MOTIF_CONSENSUS_COL = "Consensus"
+MOTIF_ORIGINAL_NAME_COL = 'Motif_Name_Original'
 
-SCAN_SCORE_COL = "Tomat0_Score"
+SCAN_SCORE_COL = "Inferelator_Score"
 SCORE_PER_BASE = "Per Base Array"
 
+DEGEN_LOOKUP = {frozenset(("A", "T")): "W",
+                frozenset(("A", "C")): "M",
+                frozenset(("A", "G")): "R",
+                frozenset(("C", "G")): "S",
+                frozenset(("C", "T")): "Y",
+                frozenset(("G", "T")): "K",
+                frozenset("A"): "A",
+                frozenset("T"): "T",
+                frozenset("G"): "G",
+                frozenset("C"): "C"}
+
 
 class Motif:
     motif_id = None
@@ -36,6 +51,7 @@ class Motif:
     _motif_accession = None
     _alphabet_map = None
     _consensus_seq = None
+    _consensus_seq_degen = None
     _info_matrix = None
     _homer_odds = None
 
@@ -46,7 +62,7 @@ class Motif:
     @alphabet.setter
     def alphabet(self, new_alphabet):
         if new_alphabet is not None:
-            self._motif_alphabet = new_alphabet
+            self._motif_alphabet = np.array(new_alphabet)
             self._alphabet_map = {ch.lower(): i for i, ch in enumerate(self._motif_alphabet)}
 
     @property
@@ -158,6 +174,16 @@ class Motif:
                                                               arr=self.probability_matrix))
         return self._consensus_seq
 
+    @property
+    def consensus_degen(self):
+        if self._consensus_seq_degen is None:
+            def _csdegen(x):
+                return DEGEN_LOOKUP[frozenset(self.alphabet[x >= 0.35])] if np.sum(x >= 0.35) > 0 else "N"
+
+            self._consensus_seq_degen = "".join(np.apply_along_axis(_csdegen, axis=1, arr=self.probability_matrix))
+
+        return self._consensus_seq_degen
+
     @property
     def max_ln_odds(self):
         max_ln_odd = np.log(np.amax(self.probability_matrix, axis=1) / 0.25)
@@ -194,10 +220,24 @@ class Motif:
                                                                    el=len(self),
                                                                    ic=self.information_content)
 
+    def __eq__(self, other):
+        try:
+            return np.allclose(self.probability_matrix, other.probability_matrix) \
+                   and (self.motif_id == other.motif_id) and (self.motif_name == other.motif_name)
+        except AttributeError:
+            pass
+
+        try:
+            return self.motif_name == other
+        except TypeError:
+            pass
+
+        return False
+
     def __init__(self, motif_id=None, motif_name=None, motif_alphabet=None, motif_background=None):
         self.id = motif_id
         self.name = motif_name
-        self.alphabet = motif_alphabet
+        self.alphabet = np.array(motif_alphabet) if motif_alphabet is not None else None
         self._motif_background = motif_background
         self._motif_probs = []
 
@@ -251,8 +291,18 @@ class Motif:
         else:
             return False
 
+    def shuffle(self, rng=None, random_seed=42):
+        """
+        Shuffles per-base probabilities
+        """
 
-class __MotifScanner:
+        if rng is not None:
+            rng.shuffle(self.probability_matrix.T)
+        else:
+            np.random.default_rng(random_seed).shuffle(self.probability_matrix.T)
+
+
+class MotifScanner:
     scanner_name = None
 
     def __init__(self, motif_file=None, motifs=None, num_workers=4):
@@ -264,58 +314,33 @@ class __MotifScanner:
         self.motifs = motifs
         self.num_workers = num_workers
 
-    def scan(self, genome_fasta_file, constraint_bed_file=None, promoter_bed=None, min_ic=None, threshold=None,
-             valid_fasta_chromosomes=None, debug=False):
+    def scan(self, genome_fasta_file=None, constraint_bed_file=None, promoter_bed=None, min_ic=None, threshold=None,
+             valid_fasta_chromosomes=None, debug=False, extracted_genome=None):
         """
         """
 
-        if valid_fasta_chromosomes is None:
-            _chr_lens = get_fasta_lengths(genome_fasta_file)
-            valid_fasta_chromosomes = list(_chr_lens.keys())
-
         # Preprocess motifs into a list of temp chunk files
         motif_files = self._preprocess(min_ic=min_ic)
         # Unpack list to a dict for convenience
         self.motifs = {mot.motif_id: mot for mot in self.motifs}
 
         try:
-            con_bed_file = load_bed_to_bedtools(constraint_bed_file) if constraint_bed_file is not None else None
-            pro_bed_file = load_bed_to_bedtools(promoter_bed) if promoter_bed is not None else None
+            if extracted_genome is None:
 
-            if con_bed_file is not None and valid_fasta_chromosomes is not None:
-                check_chromosomes_match(con_bed_file.to_dataframe(), valid_fasta_chromosomes,
-                                        chromosome_column=BED_CHROMOSOME, file_name=constraint_bed_file,
-                                        raise_no_overlap=False)
-                if debug:
-                    self._print_bed_summary(con_bed_file, constraint_bed_file)
-
-            if pro_bed_file is not None and valid_fasta_chromosomes is not None:
-                check_chromosomes_match(pro_bed_file.to_dataframe(), valid_fasta_chromosomes,
-                                        chromosome_column=BED_CHROMOSOME, file_name=pro_bed_file,
-                                        raise_no_overlap=False)
-                if debug:
-                    self._print_bed_summary(pro_bed_file, promoter_bed)
-
-            if con_bed_file is not None and pro_bed_file is not None:
-                bed_file = intersect_bed(load_bed_to_bedtools(constraint_bed_file), load_bed_to_bedtools(promoter_bed))
-            elif con_bed_file is not None:
-                bed_file = con_bed_file
-            elif pro_bed_file is not None:
-                bed_file = pro_bed_file
-            else:
-                motif_data = self._scan_extract(motif_files, genome_fasta_file, threshold=threshold)
-                return self._postprocess(motif_data)
-
-            extracted_fasta_file = extract_bed_sequence(bed_file, genome_fasta_file)
-
-            try:
-                motif_data = self._scan_extract(motif_files, extracted_fasta_file, threshold=threshold)
-                return self._postprocess(motif_data)
-            finally:
+                extracted_fasta_file = self.extract_genome(genome_fasta_file, constraint_bed_file, promoter_bed,
+                                                           valid_fasta_chromosomes, debug)
                 try:
-                    os.remove(extracted_fasta_file)
-                except FileNotFoundError:
-                    pass
+                    motif_data = self._scan_extract(motif_files, extracted_fasta_file, threshold=threshold)
+                    return self._postprocess(motif_data)
+                finally:
+                    try:
+                        os.remove(extracted_fasta_file)
+                    except FileNotFoundError:
+                        pass
+
+            else:
+                motif_data = self._scan_extract(motif_files, extracted_genome, threshold=threshold)
+                return self._postprocess(motif_data)
 
         finally:
             for file in motif_files:
@@ -324,12 +349,51 @@ class __MotifScanner:
                 except FileNotFoundError:
                     pass
 
-    def _scan_extract(self, motif_files, extracted_fasta_file, threshold=None):
+    @staticmethod
+    def extract_genome(genome_fasta_file, constraint_bed_file=None, promoter_bed=None,
+                       valid_fasta_chromosomes=None, debug=False):
+
+        if valid_fasta_chromosomes is None:
+            _chr_lens = get_fasta_lengths(genome_fasta_file)
+            valid_fasta_chromosomes = list(_chr_lens.keys())
+
+        con_bed_file = load_bed_to_bedtools(constraint_bed_file) if constraint_bed_file is not None else None
+        pro_bed_file = load_bed_to_bedtools(promoter_bed) if promoter_bed is not None else None
+
+        if con_bed_file is not None and valid_fasta_chromosomes is not None:
+            check_chromosomes_match(con_bed_file.to_dataframe(), valid_fasta_chromosomes,
+                                    chromosome_column=BED_CHROMOSOME, file_name=constraint_bed_file)
+
+            if debug:
+                MotifScanner._print_bed_summary(con_bed_file, constraint_bed_file)
+
+        if pro_bed_file is not None and valid_fasta_chromosomes is not None:
+            check_chromosomes_match(pro_bed_file.to_dataframe(), valid_fasta_chromosomes,
+                                    chromosome_column=BED_CHROMOSOME, file_name=pro_bed_file)
+
+            if debug:
+                MotifScanner._print_bed_summary(pro_bed_file, promoter_bed)
+
+        if con_bed_file is not None and pro_bed_file is not None:
+            bed_file = intersect_bed(load_bed_to_bedtools(constraint_bed_file), load_bed_to_bedtools(promoter_bed))
+        elif con_bed_file is not None:
+            bed_file = con_bed_file
+        elif pro_bed_file is not None:
+            bed_file = pro_bed_file
+        else:
+            extracted_fasta_file = tempfile.mkstemp(suffix=".fasta")[1]
+            shutil.copy2(genome_fasta_file, extracted_fasta_file)
+            return extracted_fasta_file
+
+        extracted_fasta_file = extract_bed_sequence(bed_file, genome_fasta_file)
+        return extracted_fasta_file
+
+    def _scan_extract(self, motif_files, extracted_fasta_file, threshold=None, parse_genomic_coord=True):
         # If the number of workers is 1, run fimo directly
         if (self.num_workers == 1) or (len(motif_files) == 1):
             assert len(motif_files) == 1
-            print("Launching {name} scanner [1 / 1]".format(name=self.scanner_name))
-            return self._get_motifs(extracted_fasta_file, motif_files[0], threshold=threshold)
+            return self._get_motifs(extracted_fasta_file, motif_files[0], threshold=threshold,
+                                    parse_genomic_coord=parse_genomic_coord)
 
         # Otherwise parallelize with a process pool (pathos because dill will do local functions)
         else:
@@ -338,7 +402,8 @@ class __MotifScanner:
 
             def _get_chunk_motifs(i, chunk_file):
                 print("Launching {name} scanner [{i} / {n}]".format(name=self.scanner_name, i=i + 1, n=n))
-                results = self._get_motifs(extracted_fasta_file, chunk_file, threshold=threshold)
+                results = self._get_motifs(extracted_fasta_file, chunk_file, threshold=threshold,
+                                           parse_genomic_coord=parse_genomic_coord)
                 print("Scanning completed [{i} / {n}]".format(i=i + 1, n=n))
                 return results
 
@@ -354,7 +419,7 @@ class __MotifScanner:
     def _postprocess(self, motif_peaks):
         raise NotImplementedError
 
-    def _get_motifs(self, fasta_file, motif_file, threshold=None):
+    def _get_motifs(self, fasta_file, motif_file, threshold=None, parse_genomic_coord=True):
         raise NotImplementedError
 
     def _parse_output(self, output_handle):
@@ -373,11 +438,13 @@ def motifs_to_dataframe(motifs):
     info = list(map(lambda x: x.information_content, motifs))
     ids = list(map(lambda x: x.motif_id, motifs))
     names = list(map(lambda x: x.motif_name, motifs))
+    cons = list(map(lambda x: x.consensus_degen, motifs))
 
     df = pd.DataFrame(
-        [ids, names, info, entropy, occurrence, list(map(lambda x: len(x), motifs))],
-        columns=list(map(lambda x: x.motif_name, motifs)),
-        index=[MOTIF_COL, MOTIF_NAME_COL, INFO_COL, ENTROPY_COL, OCC_COL, LEN_COL]).T
+        zip(ids, names, info, entropy, occurrence, list(map(lambda x: len(x), motifs)), motifs, cons),
+        index=list(map(lambda x: x.motif_name, motifs)),
+        columns=[MOTIF_COL, MOTIF_NAME_COL, INFO_COL, ENTROPY_COL, OCC_COL, LEN_COL, MOTIF_OBJ_COL, MOTIF_CONSENSUS_COL]
+    )
 
     return df
 
@@ -432,6 +499,63 @@ def truncate_motifs(motifs, truncate_value):
         [x.truncate(threshold=truncate_value) for x in motifs]
 
 
+def shuffle_motifs(motifs, random_seed):
+    """
+    Shuffle probabilities for each base as a random control
+
+    :param motifs: List of motif objects
+    :type motifs: list[Motif]
+    :param random_seed: Random generator seed
+    :type random_seed: int
+    """
+
+    [m.shuffle(rng=np.random.default_rng(random_seed)) for m in motifs]
+
+
+def fuzzy_merge_motifs(motif_dataframe, merge_col=MOTIF_NAME_COL, remove_dimers=False):
+
+    motif_dataframe = motif_dataframe.copy()
+    motif_dataframe[MOTIF_ORIGINAL_NAME_COL] = motif_dataframe[MOTIF_NAME_COL].copy()
+
+    motif_dataframe['FUZZY_MATCH'] = motif_dataframe[merge_col].str.lower()
+
+    # Drop any dashes (e.g ETS-1 ETS1 should merge)
+    motif_dataframe['FUZZY_MATCH'] = motif_dataframe['FUZZY_MATCH'].str.replace("-", "", regex=False)
+
+    # Drop anything in parens (e.g. Tal-1 (Scl) should just be Tal-1)
+    motif_dataframe['FUZZY_MATCH'] = motif_dataframe['FUZZY_MATCH'].str.replace("\(.*\)", "", regex=True)
+
+    # Polish up the ends to get rid of crap
+    motif_dataframe['FUZZY_MATCH'] = motif_dataframe['FUZZY_MATCH'].str.strip(",-:\\/ \r\n")
+
+    matched_data = []
+    for motif_fuzz, fuzz_df in motif_dataframe.groupby('FUZZY_MATCH'):
+        if len(fuzz_df) > 1:
+            _nn = fuzz_df[MOTIF_NAME_COL].value_counts().index[0]
+        else:
+            _nn = fuzz_df[MOTIF_NAME_COL].iloc[0]
+
+        def _renamer(obj):
+            obj.motif_name = _nn
+
+        fuzz_df['Old_Motif_Name'] = fuzz_df[MOTIF_NAME_COL]
+        fuzz_df[MOTIF_OBJ_COL].apply(_renamer)
+        fuzz_df[MOTIF_NAME_COL] = _nn
+
+        matched_data.append(fuzz_df)
+
+    matched_data = pd.concat(matched_data)
+
+    if remove_dimers:
+        matched_data = matched_data.loc[~matched_data[MOTIF_NAME_COL].str.contains(":"), :]
+        matched_data = matched_data.loc[~matched_data[MOTIF_NAME_COL].str.contains(", "), :]
+        matched_data = matched_data.loc[~matched_data[MOTIF_NAME_COL].str.contains(";"), :].copy()
+
+    print("Fuzzy merging completed: {p} TFs reduced to {n} TFs".format(n=len(matched_data[MOTIF_NAME_COL].unique()),
+                                                                       p=len(motif_dataframe[MOTIF_NAME_COL].unique())))
+    return matched_data
+
+
 def chunk_motifs(file_type, motifs, num_workers=4, min_ic=None):
     """
     Break a motif file up into chunks
@@ -454,8 +578,9 @@ def chunk_motifs(file_type, motifs, num_workers=4, min_ic=None):
         motifs = list(itertools.compress(motifs, [m.information_content >= min_ic for m in motifs]))
 
     if num_workers == 1:
-        file_name = os.path.join(temp_dir, "chunk1.mchunk")
-        file_type.write(file_name, motifs)
+        tf_h, file_name = tempfile.mkstemp(prefix='1', suffix=".mchunk", dir=temp_dir)
+        with os.fdopen(tf_h, "w", 1) as tf_fh:
+            file_type.write(tf_fh, motifs)
         return [file_name]
 
     num_workers = len(motifs) if num_workers > len(motifs) else num_workers
@@ -465,8 +590,9 @@ def chunk_motifs(file_type, motifs, num_workers=4, min_ic=None):
     files = []
 
     for i in range(num_workers):
-        file_name = os.path.join(temp_dir, "chunk" + str(i) + ".mchunk")
-        file_type.write(file_name, [m for m, b in zip(motifs, (chunk_index == i)) if b])
+        tf_h, file_name = tempfile.mkstemp(prefix=str(i), suffix=".mchunk", dir=temp_dir)
+        with os.fdopen(tf_h, "w", 1) as tf_fh:
+            file_type.write(tf_fh, [m for m, b in zip(motifs, (chunk_index == i)) if b])
         files.append(file_name)
 
     return files
