@@ -1,7 +1,9 @@
-import pybedtools
+import gzip
 import pandas as pd
+import re
 
-GENE_ID_REGEX = 'gene_id\s\"([A-Za-z0-9\.\-\(\)]+)\"\;'
+GTF_GENE_ID_REGEX = 'gene_id\s\"(.*?)\"\;'
+GFF_GENE_ID_REGEX = '\=gene:(.*?)[;|$]'
 
 # Column names
 GTF_ATTRIBUTES = 'attributes'
@@ -12,13 +14,17 @@ SEQ_START = 'start'
 SEQ_STOP = 'end'
 SEQ_TSS = 'TSS'
 
+WINDOW_UP = "Window_Start"
+WINDOW_DOWN = "Window_End"
+
 GTF_COLUMNS = ["seqname", "source", "feature", "start", "end", "score", "strand", "frame", "attributes"]
 
 
 def load_gtf_to_dataframe(gtf_path, fasta_record_lengths=None):
     """
-    Loads genes from a GTF into a dataframe and returns them
-    :param gtf_path: Path to the GTF file
+    Loads genes from a GTF or GFF into a dataframe and returns them
+
+    :param gtf_path: Path to the GTF or GFF file
     :type gtf_path: str
     :param fasta_record_lengths: A dict of valid FASTA records, keyed by name
     :type fasta_record_lengths: dict(int)
@@ -31,9 +37,17 @@ def load_gtf_to_dataframe(gtf_path, fasta_record_lengths=None):
     :rtype: pd.DataFrame
     """
 
+    if any(gtf_path.lower().endswith(x) for x in [".gff", ".gff3", ".gff.gz", ".gff3.gz"]):
+        _gregex = GFF_GENE_ID_REGEX
+    else:
+        _gregex = GTF_GENE_ID_REGEX
+
     # Load annotations into a dataframe with pybedtools
     annotations = pd.read_csv(gtf_path, sep="\t", names=GTF_COLUMNS, comment="#")
 
+    if len(annotations) == 0:
+        raise ValueError("No records present in {f}".format(f=gtf_path))
+
     # Fix chromosome names to always be strings
     annotations[GTF_CHROMOSOME] = annotations[GTF_CHROMOSOME].astype(str)
 
@@ -45,7 +59,13 @@ def load_gtf_to_dataframe(gtf_path, fasta_record_lengths=None):
     annotations = annotations.loc[~pd.isnull(annotations[SEQ_START]) & ~pd.isnull(annotations[SEQ_STOP]), :]
 
     # Regex extract the gene_id from the annotations column
-    annotations[GTF_GENENAME] = annotations[GTF_ATTRIBUTES].str.extract(GENE_ID_REGEX, expand=False)
+    annotations[GTF_GENENAME] = annotations[GTF_ATTRIBUTES].str.extract(_gregex, expand=False, flags=re.IGNORECASE)
+
+    # Drop any NaNs in GENE_NAME:
+    annotations.dropna(inplace=True, subset=[GTF_GENENAME])
+
+    if len(annotations) == 0:
+        raise ValueError("Unable to parse gene IDs from annotation file attributes")
 
     # Define genes as going from the minimum start for any subfeature to the maximum end for any subfeature
     annotations = _fix_genes(annotations)
@@ -53,17 +73,21 @@ def load_gtf_to_dataframe(gtf_path, fasta_record_lengths=None):
     return _add_TSS(annotations)
 
 
-def open_window(annotation_dataframe, window_size, use_tss=False, fasta_record_lengths=None):
+def open_window(annotation_dataframe, window_size, use_tss=False, fasta_record_lengths=None,
+                constrain_to_intergenic=False):
     """
     This needs to adjust the start and stop in the annotation dataframe with window sizes
+
     :param annotation_dataframe: pd.DataFrame
     :param window_size: int
     :param use_tss: bool
     :param fasta_record_lengths:
     :return window_annotate: pd.DataFrame
     """
+    
     window_annotate = annotation_dataframe.copy()
-
+    window_annotate[WINDOW_UP], window_annotate[WINDOW_DOWN] = pd.NA, pd.NA
+    
     try:
         if len(window_size) == 1:
             w_up, w_down = window_size[0], window_size[0]
@@ -75,17 +99,17 @@ def open_window(annotation_dataframe, window_size, use_tss=False, fasta_record_l
         w_up, w_down = window_size, window_size
 
     if use_tss:
-        window_annotate.loc[window_annotate[GTF_STRAND] == "+", SEQ_START] = window_annotate[SEQ_TSS] - w_up
-        window_annotate.loc[window_annotate[GTF_STRAND] == "+", SEQ_STOP] = window_annotate[SEQ_TSS] + w_down
-        window_annotate.loc[window_annotate[GTF_STRAND] == "-", SEQ_START] = window_annotate[SEQ_TSS] - w_down
-        window_annotate.loc[window_annotate[GTF_STRAND] == "-", SEQ_STOP] = window_annotate[SEQ_TSS] + w_up
+        window_annotate.loc[window_annotate[GTF_STRAND] == "+", WINDOW_UP] = window_annotate[SEQ_TSS] - w_up
+        window_annotate.loc[window_annotate[GTF_STRAND] == "+", WINDOW_DOWN] = window_annotate[SEQ_TSS] + w_down
+        window_annotate.loc[window_annotate[GTF_STRAND] == "-", WINDOW_UP] = window_annotate[SEQ_TSS] - w_down
+        window_annotate.loc[window_annotate[GTF_STRAND] == "-", WINDOW_DOWN] = window_annotate[SEQ_TSS] + w_up
     else:
-        window_annotate.loc[window_annotate[GTF_STRAND] == "+", SEQ_START] = window_annotate[SEQ_START] - w_up
-        window_annotate.loc[window_annotate[GTF_STRAND] == "+", SEQ_STOP] = window_annotate[SEQ_STOP] + w_down
-        window_annotate.loc[window_annotate[GTF_STRAND] == "-", SEQ_START] = window_annotate[SEQ_START] - w_down
-        window_annotate.loc[window_annotate[GTF_STRAND] == "-", SEQ_STOP] = window_annotate[SEQ_STOP] + w_up
+        window_annotate.loc[window_annotate[GTF_STRAND] == "+", WINDOW_UP] = window_annotate[SEQ_START] - w_up
+        window_annotate.loc[window_annotate[GTF_STRAND] == "+", WINDOW_DOWN] = window_annotate[SEQ_STOP] + w_down
+        window_annotate.loc[window_annotate[GTF_STRAND] == "-", WINDOW_UP] = window_annotate[SEQ_START] - w_down
+        window_annotate.loc[window_annotate[GTF_STRAND] == "-", WINDOW_DOWN] = window_annotate[SEQ_STOP] + w_up
 
-    window_annotate.loc[window_annotate[SEQ_START] < 1, SEQ_START] = 1
+    window_annotate.loc[window_annotate[WINDOW_UP] < 1, WINDOW_UP] = 1
 
     if fasta_record_lengths is not None:
 
@@ -94,11 +118,42 @@ def open_window(annotation_dataframe, window_size, use_tss=False, fasta_record_l
         for chromosome in _gtf_fasta_match:
             _chrlen = fasta_record_lengths[chromosome]
             _idx = window_annotate[GTF_CHROMOSOME] == chromosome
-            window_annotate.loc[_idx & (window_annotate[SEQ_STOP] > _chrlen), SEQ_STOP] = _chrlen
-            window_annotate.loc[_idx & (window_annotate[SEQ_START] > _chrlen), SEQ_START] = _chrlen
+            window_annotate.loc[_idx & (window_annotate[WINDOW_UP] > _chrlen), WINDOW_UP] = _chrlen
+            window_annotate.loc[_idx & (window_annotate[WINDOW_DOWN] > _chrlen), WINDOW_DOWN] = _chrlen       
+
+    if constrain_to_intergenic:
+        window_annotate = window_annotate.groupby(GTF_CHROMOSOME).apply(fix_overlap)
+        window_annotate.reset_index(level=GTF_CHROMOSOME, inplace=True, drop=True)
+
+    window_annotate[SEQ_START] = window_annotate[WINDOW_UP]
+    window_annotate[SEQ_STOP] = window_annotate[WINDOW_DOWN]
+    window_annotate.drop([WINDOW_UP, WINDOW_DOWN], axis=1, inplace=True)
 
     return window_annotate
 
+def fix_overlap(dataframe):
+    """
+    Apply function that sets window start and stop positions so that they do not overlap with features
+
+    :param dataframe: pd.DataFrame
+    :return dataframe: pd.DataFrame
+    """
+    
+    dataframe = dataframe.sort_values(by=SEQ_START)
+    windows = dataframe[[WINDOW_UP, WINDOW_DOWN]].copy()
+    
+    start_idx = dataframe[WINDOW_UP] < dataframe[SEQ_STOP].shift(1)
+    dataframe.loc[start_idx, WINDOW_UP] = dataframe[SEQ_STOP].shift(1).loc[start_idx].astype(int)
+
+    stop_idx = dataframe[WINDOW_DOWN] > dataframe[SEQ_START].shift(-1)
+    dataframe.loc[stop_idx, WINDOW_DOWN] = dataframe[SEQ_START].shift(-1).loc[stop_idx].astype(int)
+
+    # Undo checking for intergenic when there's a problem - usually it means overlapping genes, which I can't deal with
+    bad_idx = dataframe[WINDOW_UP] >= dataframe[WINDOW_DOWN]
+    dataframe.loc[bad_idx, [WINDOW_UP, WINDOW_DOWN]] = windows.loc[bad_idx, [WINDOW_UP, WINDOW_DOWN]]
+
+    return dataframe
+
 
 def get_fasta_lengths(fasta_file):
     """
@@ -110,7 +165,10 @@ def get_fasta_lengths(fasta_file):
     """
 
     fasta_len = {}
-    with open(fasta_file, mode="r") as fasta_fh:
+
+    _opener = gzip.open if fasta_file.endswith(".gz") else open
+
+    with _opener(fasta_file, mode="rt") as fasta_fh:
         current_record = None
         for line in fasta_fh:
             if line.startswith(">"):
@@ -208,7 +266,7 @@ def _fix_genes(gene_dataframe):
     # Define the functions for aggregating gene records
     aggregate_functions = {SEQ_START: min, SEQ_STOP: max, GTF_CHROMOSOME: _most_common, GTF_STRAND: _most_common}
 
-    return gene_dataframe.groupby("gene_name").aggregate(aggregate_functions).reset_index()
+    return gene_dataframe.groupby(GTF_GENENAME).aggregate(aggregate_functions).reset_index()
 
 
 def _add_TSS(gene_dataframe):
